@@ -33,12 +33,13 @@
 #include <linux/if_ether.h>
 #include <arpa/inet.h>
 #include <linux/filter.h>
+
 #include <ell/ell.h>
 
-#include "crypto.h"
-#include "ie.h"
-#include "util.h"
-#include "handshake.h"
+#include "src/crypto.h"
+#include "src/ie.h"
+#include "src/util.h"
+#include "src/handshake.h"
 
 static bool handshake_get_nonce(uint8_t nonce[])
 {
@@ -72,10 +73,10 @@ void __handshake_set_install_igtk_func(handshake_install_igtk_func_t func)
 
 void handshake_state_free(struct handshake_state *s)
 {
-	typeof(s->free) destroy = s->free;
+	__typeof__(s->free) destroy = s->free;
 
-	l_free(s->ap_ie);
-	l_free(s->own_ie);
+	l_free(s->authenticator_ie);
+	l_free(s->supplicant_ie);
 	l_free(s->mde);
 	l_free(s->fte);
 	l_free(s->passphrase);
@@ -98,6 +99,11 @@ void handshake_state_set_authenticator_address(struct handshake_state *s,
 	memcpy(s->aa, aa, sizeof(s->aa));
 }
 
+void handshake_state_set_authenticator(struct handshake_state *s, bool auth)
+{
+	s->authenticator = auth;
+}
+
 void handshake_state_set_pmk(struct handshake_state *s, const uint8_t *pmk,
 				size_t pmk_len)
 {
@@ -116,28 +122,6 @@ struct l_settings *handshake_state_get_8021x_config(struct handshake_state *s)
 	return s->settings_8021x;
 }
 
-static void handshake_state_set_ap_ie(struct handshake_state *s,
-					const uint8_t *ie, bool is_wpa)
-{
-	l_free(s->ap_ie);
-	s->ap_ie = l_memdup(ie, ie[1] + 2u);
-	s->wpa_ie = is_wpa;
-}
-
-static void handshake_state_set_own_ie(struct handshake_state *s,
-					const uint8_t *ie, bool is_wpa)
-{
-	l_free(s->own_ie);
-	s->own_ie = l_memdup(ie, ie[1] + 2u);
-	s->wpa_ie = is_wpa;
-}
-
-void handshake_state_set_ap_rsn(struct handshake_state *s,
-				const uint8_t *rsn_ie)
-{
-	handshake_state_set_ap_ie(s, rsn_ie, false);
-}
-
 static bool handshake_state_setup_own_ciphers(struct handshake_state *s,
 						const struct ie_rsn_info *info)
 {
@@ -156,36 +140,74 @@ static bool handshake_state_setup_own_ciphers(struct handshake_state *s,
 	return true;
 }
 
-bool handshake_state_set_own_rsn(struct handshake_state *s,
+static bool handshake_state_set_authenticator_ie(struct handshake_state *s,
+						const uint8_t *ie, bool is_wpa)
+{
+	struct ie_rsn_info info;
+
+	l_free(s->authenticator_ie);
+	s->authenticator_ie = l_memdup(ie, ie[1] + 2u);
+	s->wpa_ie = is_wpa;
+
+	if (!s->authenticator)
+		return true;
+
+	if (is_wpa) {
+		if (ie_parse_wpa_from_data(ie, ie[1] + 2, &info) < 0)
+			return false;
+	} else {
+		if (ie_parse_rsne_from_data(ie, ie[1] + 2, &info) < 0)
+			return false;
+	}
+
+	return handshake_state_setup_own_ciphers(s, &info);
+}
+
+static bool handshake_state_set_supplicant_ie(struct handshake_state *s,
+						const uint8_t *ie, bool is_wpa)
+{
+	struct ie_rsn_info info;
+
+	l_free(s->supplicant_ie);
+	s->supplicant_ie = l_memdup(ie, ie[1] + 2u);
+	s->wpa_ie = is_wpa;
+
+	if (s->authenticator)
+		return true;
+
+	if (is_wpa) {
+		if (ie_parse_wpa_from_data(ie, ie[1] + 2, &info) < 0)
+			return false;
+	} else {
+		if (ie_parse_rsne_from_data(ie, ie[1] + 2, &info) < 0)
+			return false;
+	}
+
+	return handshake_state_setup_own_ciphers(s, &info);
+}
+
+bool handshake_state_set_authenticator_rsn(struct handshake_state *s,
+						const uint8_t *rsn_ie)
+{
+	return handshake_state_set_authenticator_ie(s, rsn_ie, false);
+}
+
+bool handshake_state_set_supplicant_rsn(struct handshake_state *s,
 					const uint8_t *rsn_ie)
 {
-	struct ie_rsn_info info;
-
-	handshake_state_set_own_ie(s, rsn_ie, false);
-
-	if (ie_parse_rsne_from_data(rsn_ie, rsn_ie[1] + 2, &info) < 0)
-		return false;
-
-	return handshake_state_setup_own_ciphers(s, &info);
+	return handshake_state_set_supplicant_ie(s, rsn_ie, false);
 }
 
-void handshake_state_set_ap_wpa(struct handshake_state *s,
+bool handshake_state_set_authenticator_wpa(struct handshake_state *s,
 				const uint8_t *wpa_ie)
 {
-	handshake_state_set_ap_ie(s, wpa_ie, true);
+	return handshake_state_set_authenticator_ie(s, wpa_ie, true);
 }
 
-bool handshake_state_set_own_wpa(struct handshake_state *s,
+bool handshake_state_set_supplicant_wpa(struct handshake_state *s,
 					const uint8_t *wpa_ie)
 {
-	struct ie_rsn_info info;
-
-	handshake_state_set_own_ie(s, wpa_ie, true);
-
-	if (ie_parse_wpa_from_data(wpa_ie, wpa_ie[1] + 2, &info) < 0)
-		return false;
-
-	return handshake_state_setup_own_ciphers(s, &info);
+	return handshake_state_set_supplicant_ie(s, wpa_ie, true);
 }
 
 void handshake_state_set_ssid(struct handshake_state *s, const uint8_t *ssid,
@@ -424,6 +446,21 @@ bool handshake_state_get_pmkid(struct handshake_state *s, uint8_t *out_pmkid)
 					use_sha256);
 }
 
+void handshake_state_set_gtk(struct handshake_state *s, const uint8_t *key,
+				unsigned int key_index, const uint8_t *rsc)
+{
+	enum crypto_cipher cipher =
+		ie_rsn_cipher_suite_to_cipher(s->group_cipher);
+	int key_len = crypto_cipher_key_len(cipher);
+
+	if (!key_len)
+		return;
+
+	memcpy(s->gtk, key, key_len);
+	s->gtk_index = key_index;
+	memcpy(s->gtk_rsc, rsc, 6);
+}
+
 /*
  * This function performs a match of the RSN/WPA IE obtained from the scan
  * results vs the RSN/WPA IE obtained as part of the 4-way handshake.  If they
@@ -517,7 +554,7 @@ bool handshake_util_ap_ie_matches(const uint8_t *msg_ie,
 }
 
 static const uint8_t *find_kde(const uint8_t *data, size_t data_len,
-				size_t *out_len, const unsigned char *oui)
+				size_t *out_len, enum handshake_kde selector)
 {
 	struct ie_tlv_iter iter;
 	const uint8_t *result;
@@ -535,7 +572,7 @@ static const uint8_t *find_kde(const uint8_t *data, size_t data_len,
 
 		/* Check OUI */
 		result = ie_tlv_iter_get_data(&iter);
-		if (memcmp(result, oui, 4))
+		if (l_get_be32(result) != selector)
 			continue;
 
 		if (out_len)
@@ -550,9 +587,9 @@ static const uint8_t *find_kde(const uint8_t *data, size_t data_len,
 const uint8_t *handshake_util_find_gtk_kde(const uint8_t *data, size_t data_len,
 						size_t *out_gtk_len)
 {
-	static const unsigned char gtk_oui[] = { 0x00, 0x0f, 0xac, 0x01 };
 	size_t gtk_len;
-	const uint8_t *gtk = find_kde(data, data_len, &gtk_len, gtk_oui);
+	const uint8_t *gtk = find_kde(data, data_len, &gtk_len,
+					HANDSHAKE_KDE_GTK);
 
 	if (!gtk)
 		return NULL;
@@ -577,9 +614,9 @@ const uint8_t *handshake_util_find_igtk_kde(const uint8_t *data,
 						size_t data_len,
 						size_t *out_igtk_len)
 {
-	static const unsigned char igtk_oui[] = { 0x00, 0x0f, 0xac, 0x09 };
 	size_t igtk_len;
-	const uint8_t *igtk = find_kde(data, data_len, &igtk_len, igtk_oui);
+	const uint8_t *igtk = find_kde(data, data_len, &igtk_len,
+					HANDSHAKE_KDE_IGTK);
 
 	if (!igtk)
 		return NULL;
@@ -603,11 +640,10 @@ const uint8_t *handshake_util_find_igtk_kde(const uint8_t *data,
 const uint8_t *handshake_util_find_pmkid_kde(const uint8_t *data,
 						size_t data_len)
 {
-	static const unsigned char pmkid_oui[] = { 0x00, 0x0f, 0xac, 0x04 };
 	const uint8_t *pmkid;
 	size_t pmkid_len;
 
-	pmkid = find_kde(data, data_len, &pmkid_len, pmkid_oui);
+	pmkid = find_kde(data, data_len, &pmkid_len, HANDSHAKE_KDE_PMKID);
 
 	if (pmkid && pmkid_len != 16)
 		return NULL;
@@ -615,6 +651,21 @@ const uint8_t *handshake_util_find_pmkid_kde(const uint8_t *data,
 	return pmkid;
 }
 
+/* Defined in 802.11-2016 12.7.2 j), Figure 12-34 */
+void handshake_util_build_gtk_kde(enum crypto_cipher cipher, const uint8_t *key,
+					unsigned int key_index, uint8_t *to)
+{
+	size_t key_len = crypto_cipher_key_len(cipher);
+
+	*to++ = IE_TYPE_VENDOR_SPECIFIC;
+	*to++ = 6 + key_len;
+	l_put_be32(HANDSHAKE_KDE_GTK, to);
+	to += 4;
+	*to++ = key_index;
+	*to++ = 0;
+	memcpy(to, key, key_len);
+}
+
 /*
  * Unwrap a GTK / IGTK included in an FTE following 802.11-2012, Section 12.8.5:
  *
