@@ -488,28 +488,36 @@ void network_sync_psk(struct network *network)
 					network->settings);
 }
 
-static bool bss_is_sae(struct scan_bss *bss)
+static inline bool __bss_is_sae(const struct scan_bss *bss,
+						const struct ie_rsn_info *rsn)
+{
+	if (rsn->akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
+		return true;
+
+	return false;
+}
+
+static bool bss_is_sae(const struct scan_bss *bss)
 {
 	struct ie_rsn_info rsn;
 
 	memset(&rsn, 0, sizeof(rsn));
 	scan_bss_get_rsn_info(bss, &rsn);
 
-	if (rsn.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
-		return true;
-
-	return false;
+	return __bss_is_sae(bss, &rsn);
 }
 
 int network_autoconnect(struct network *network, struct scan_bss *bss)
 {
 	struct station *station = network->station;
 	struct wiphy *wiphy = station_get_wiphy(station);
+	enum security security = network_get_security(network);
+	struct ie_rsn_info rsn;
 	bool is_autoconnectable;
 	bool is_rsn;
 	int ret;
 
-	switch (network_get_security(network)) {
+	switch (security) {
 	case SECURITY_NONE:
 		is_rsn = false;
 		break;
@@ -528,24 +536,6 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 	if (!network_settings_load(network))
 		return -ENOKEY;
 
-	if (is_rsn) {
-		struct ie_rsn_info rsn;
-
-		memset(&rsn, 0, sizeof(rsn));
-		scan_bss_get_rsn_info(bss, &rsn);
-
-		if (!wiphy_select_cipher(wiphy, rsn.pairwise_ciphers) ||
-				!wiphy_select_cipher(wiphy, rsn.group_cipher)) {
-			l_debug("Cipher mis-match");
-			ret = -ENETUNREACH;
-			goto close_settings;
-		}
-
-		ret = network_load_psk(network, bss_is_sae(bss));
-		if (ret < 0)
-			goto close_settings;
-	}
-
 	/* If no entry, default to Autoconnectable=True */
 	if (!l_settings_get_bool(network->settings, "Settings",
 					"Autoconnect", &is_autoconnectable))
@@ -555,7 +545,24 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 	if (!is_autoconnectable)
 		goto close_settings;
 
-	if (network_get_security(network) == SECURITY_8021X) {
+	if (!is_rsn)
+		goto done;
+
+	memset(&rsn, 0, sizeof(rsn));
+	scan_bss_get_rsn_info(bss, &rsn);
+
+	if (!wiphy_select_cipher(wiphy, rsn.pairwise_ciphers) ||
+			!wiphy_select_cipher(wiphy, rsn.group_cipher)) {
+		l_debug("Cipher mis-match");
+		ret = -ENETUNREACH;
+		goto close_settings;
+	}
+
+	if (security == SECURITY_PSK) {
+		ret = network_load_psk(network, __bss_is_sae(bss, &rsn));
+		if (ret < 0)
+			goto close_settings;
+	} else if (security == SECURITY_8021X) {
 		struct l_queue *missing_secrets = NULL;
 
 		ret = eap_check_settings(network->settings, network->secrets,
@@ -573,6 +580,7 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 			goto close_settings;
 	}
 
+done:
 	return __station_connect_network(station, network, bss);
 
 close_settings:
